/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAB_DEBUG			FALSE

#include "surface.pmh"

namespace fe
{

sp<SurfaceRange> SurfaceAccessibleBase::atomize(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomize\n");
#endif

	if(a_atomicChange==e_pointsOfPrimitives)
	{
		return atomizeConnectivity(a_atomicChange,a_group,a_desiredCount);
	}

	//* fallback to non-atomic

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);
	const Element element=asPoints? e_point: e_primitive;
	sp<SurfaceAccessorI> spAccessor=accessor(element,e_generic);
	const I32 elementCount=spAccessor.isValid()? spAccessor->count(): 0;
	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	//* use everything

	if(a_group.empty())
	{
		if(elementCount>0)
		{
			spSurfaceRange->nonAtomic().add(0,elementCount-1);
		}
		return spSurfaceRange;
	}

	//* use group

	I32 spanStart= -1;
	SurfaceRange::MultiSpan& rMultiSpan=spSurfaceRange->nonAtomic();

	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const I32 elementPage=
				asPoints? pointPage(elementIndex): primitivePage(elementIndex);
		if(spanStart<0 && elementPage>=0)
		{
			//* begin span
			spanStart=elementIndex;
		}
		else if(spanStart>=0 && elementPage<0)
		{
			//* complete span
			rMultiSpan.add(spanStart,elementIndex-1);
			spanStart=-1;
		}
	}

	//* complete last span
	if(spanStart>=0)
	{
		rMultiSpan.add(spanStart,elementCount-1);

	}

	return spSurfaceRange;
}

sp<SurfaceRange> SurfaceAccessibleBase::atomizeConnectivity(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
	preparePaging(a_atomicChange,a_group);

	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	sp<SurfaceAccessorI> spPoints=accessor(e_point,e_position);
	sp<SurfaceAccessorI> spVertices=accessor(e_primitive,e_vertices);

	const U32 pointCount=spPoints->count();
	const U32 primitiveCount=spVertices->count();

	std::vector<U32> pointPatch(pointCount,0);
	std::vector<I32> primitivePatch(primitiveCount,0);

	std::vector<U32> pagePatch(primitiveCount+1,0);	//* pages <= prims+1
	std::vector<U32> histogram(primitiveCount,0);	//* patches <= prims

	U32 maxPatch=0;
	U32 maxPage=0;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		if(primitivePage(primitiveIndex)<0)
		{
			primitivePatch[primitiveIndex]= -1;
			continue;
		}

		U32 lowPatch=0;
		U32 lowPage=0;
		BWORD multiPatch=FALSE;
		BWORD zeroPatches=FALSE;
		BWORD multiPage=FALSE;

		const U32 subCount=spVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=spVertices->integer(primitiveIndex,subIndex);
			U32& rPatch=pointPatch[pointIndex];
			FEASSERT(rPatch<=maxPatch);

			const I32 page=pointPage(pointIndex);
			if(page>0)
			{
				const U32 existingPatch=pagePatch[page];
				FEASSERT(existingPatch<=maxPatch);

				if(existingPatch)
				{
					rPatch=existingPatch;
				}
				if(!lowPage)
				{
					lowPage=page;
				}
				else if(lowPage!=page)
				{
					multiPage=TRUE;
					if(lowPage>page)
					{
						lowPage=page;
					}
				}
			}
			if(rPatch)
			{
				if(!lowPatch)
				{
					lowPatch=rPatch;
				}
				else if(lowPatch!=rPatch)
				{
					multiPatch=TRUE;
					if(lowPatch>rPatch)
					{
						lowPatch=rPatch;
					}
				}
			}
			else
			{
				zeroPatches=TRUE;
			}
		}

		if(multiPage)
		{
			//* leave for postAtomic
			lowPatch=0;
		}
		else if(multiPatch)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				const U32 patch=pointPatch[pointIndex];
				if(patch!=lowPatch)
				{
					for(U32 pointOther=0;pointOther<pointCount;pointOther++)
					{
						U32& rOtherPatch=pointPatch[pointOther];

						if(rOtherPatch==patch)
						{
							//* merge all 'patch' into 'lowPatch'
							rOtherPatch=lowPatch;
						}
						else if(rOtherPatch==maxPatch)
						{
							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPatch=patch;
						}
					}
					FEASSERT(maxPatch);
					maxPatch--;
				}
			}
		}
		else if(!lowPatch)
		{
			maxPatch++;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				pointPatch[pointIndex]=maxPatch;
			}
			lowPatch=maxPatch;
		}
		else if(zeroPatches)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				pointPatch[pointIndex]=lowPatch;
			}
		}

		FEASSERT(lowPatch<=maxPatch);

		primitivePatch[primitiveIndex]=lowPatch;
		histogram[lowPatch]+=subCount;
		if(lowPatch)
		{
			pagePatch[lowPage]=lowPatch;
		}
		if(maxPage<lowPage)
		{
			maxPage=lowPage;
		}
	}

#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPage %d\n");
	for(U32 pageIndex=0;pageIndex<=maxPage;pageIndex++)
	{
		feLog("page %d uses patch %d\n",pageIndex,pagePatch[pageIndex]);
	}
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPatch %d\n");
	for(U32 patchIndex=0;patchIndex<=maxPatch;patchIndex++)
	{
		feLog("patch %d has %d points\n",patchIndex,histogram[patchIndex]);
	}
#endif

	const U32 desiredSize=a_desiredCount? primitiveCount/a_desiredCount: 1;

	std::vector<U32> patchAtom(maxPatch+1,0);

	I32 maxAtom=1;
	U32 atomSize=0;
	for(U32 patchIndex=1;patchIndex<=maxPatch;patchIndex++)
	{
		if(atomSize>=desiredSize)
		{
			maxAtom++;
			atomSize=0;
		}
		patchAtom[patchIndex]=maxAtom;
		atomSize+=histogram[patchIndex];
#if FE_SAB_DEBUG
		feLog("patch %d atom %d\n",patchIndex,maxAtom);
#endif
	}

	std::vector<SurfaceRange::MultiSpan*> multiSpans(maxAtom+1,NULL);
	I32 spanAtom= -1;
	I32 spanStart= -1;
#if FE_SAB_DEBUG
	feLog("first %d from %d\n",spanAtom,spanStart);
#endif
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const I32 patchIndex=primitivePatch[primitiveIndex];
		const I32 atomIndex=patchIndex>=0? patchAtom[patchIndex]: -1;

		FEASSERT(atomIndex<=maxAtom);
		FEASSERT(spanAtom<=maxAtom);
		if(atomIndex!=spanAtom)
		{
			//* complete existing span
			SurfaceRange::MultiSpan* pMultiSpan=NULL;
			if(spanAtom>=0)
			{
				FEASSERT(spanAtom<=maxAtom);
				pMultiSpan=multiSpans[spanAtom];
				if(!pMultiSpan)
				{
					pMultiSpan=spanAtom? &spSurfaceRange->addAtomic():
							&spSurfaceRange->postAtomic();
					multiSpans[spanAtom]=pMultiSpan;
				}
			}
			if(spanStart>=0)
			{
#if FE_SAB_DEBUG
				feLog("complete %d span %d %d\n",
						spanAtom,spanStart,primitiveIndex-1);
#endif
				pMultiSpan->add(spanStart,primitiveIndex-1);
				spanStart= -1;
			}

			//* start another span
			spanAtom=atomIndex;
			if(atomIndex>=0)
			{
				spanStart=primitiveIndex;

#if FE_SAB_DEBUG
				feLog("start %d from %d\n",spanAtom,spanStart);
#endif
			}
		}
	}

	//* complete last span
	SurfaceRange::MultiSpan* pMultiSpan=NULL;
	if(spanStart>=0)
	{
		FEASSERT(spanAtom<=maxAtom);
		pMultiSpan=multiSpans[spanAtom];
		if(!pMultiSpan)
		{
			pMultiSpan=spanAtom? &spSurfaceRange->addAtomic():
					&spSurfaceRange->postAtomic();
			multiSpans[spanAtom]=pMultiSpan;
		}
		pMultiSpan->add(spanStart,primitiveCount-1);
	}

#if FE_SAB_DEBUG
	feLog("brief: %s\n",spSurfaceRange->brief().c_str());
	feLog("dump:\n%s\n",spSurfaceRange->dump().c_str());
#endif

	return spSurfaceRange;
}

} // namespace
