/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_DODGE_DEBUG			FALSE
#define FE_DODGE_IMPACT_DEBUG	FALSE

#include "grass/grass.pmh"

using namespace fe;

void DodgeOp::initialize(void)
{
	catalog<String>("Threads","page")="Config";

	catalog<String>("AutoThread","page")="Config";

	catalog<String>("StayAlive","page")="Config";

	catalog<String>("Group");
	catalog<String>("Group","label")="Primitive Group";
	catalog<String>("Group","page")="Config";
	catalog<String>("Group","suggest")="primitiveGroups";

	catalog<I32>("Refinement")=0;
	catalog<I32>("Refinement","high")=4;
	catalog<I32>("Refinement","max")=8;
	catalog<String>("Refinement","page")="Config";
	catalog<String>("Refinement","hint")=
			"Maximum level of collider subdivision."
			"  Refinement may be adaptive and non-uniform.";

	catalog<I32>("Seed")=0;
	catalog<I32>("Seed","max")=RAND_MAX;
	catalog<String>("Seed","page")="Config";
	catalog<String>("Seed","hint")="Set of random results.";

	catalog<bool>("DebugOutput")=false;
	catalog<String>("DebugOutput","label")="Debug Output";
	catalog<String>("DebugOutput","page")="Config";
	catalog<String>("DebugOutput","hint")=
			"Draws lines over the output showing the interactions."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<bool>("Temporal")=false;
	catalog<String>("Temporal","page")="Behavior";
	catalog<String>("Temporal","hint")=
			"Tries to retain state of the prior frame."
			"  If the prior frame was not the last evaluated,"
			" the node may choose not to use temporal data.";
	catalog<bool>("Temporal","joined")=true;

	catalog<Real>("Response")=1.0;
	catalog<String>("Response","enabler")="Temporal";
	catalog<String>("Response","page")="Behavior";
	catalog<String>("Response","hint")=
			"When temporal, this defines how quickly the surface returns"
			" to its original shaped when released.";
	catalog<bool>("Response","joined")=true;

	catalog<Real>("ResponseRandom")=0.0;
	catalog<Real>("ResponseRandom","high")=1.0;
	catalog<String>("ResponseRandom","label")="+-";
	catalog<String>("ResponseRandom","enabler")="Temporal";
	catalog<String>("ResponseRandom","page")="Behavior";

	catalog<Real>("Threshold")=0.0;
	catalog<String>("Threshold","page")="Behavior";
	catalog<String>("Threshold","hint")="Gap to maintain around collider.";
	catalog<bool>("Threshold","joined")=true;

	catalog<Real>("ThresholdRandom")=0.0;
	catalog<Real>("ThresholdRandom","high")=1.0;
	catalog<String>("ThresholdRandom","label")="+-";
	catalog<String>("ThresholdRandom","page")="Behavior";

	catalog<Real>("Projection")=1.0;
	catalog<Real>("Projection","max")=1.0;
	catalog<String>("Projection","page")="Behavior";
	catalog<String>("Projection","hint")=
			"Amount that trailing points are affected by preceding impacts.";
	catalog<bool>("Projection","joined")=true;

	catalog<Real>("ProjectionRandom")=0.0;
	catalog<Real>("ProjectionRandom","high")=1.0;
	catalog<String>("ProjectionRandom","label")="+-";
	catalog<String>("ProjectionRandom","page")="Behavior";

	catalog<Real>("Repulsion")=0.0;
	catalog<Real>("Repulsion","high")=0.2;
	catalog<Real>("Repulsion","max")=1.0;
	catalog<String>("Repulsion","page")="Behavior";
	catalog<String>("Repulsion","hint")=
			"Push away nearby curves, even if not colliding.";
	catalog<bool>("Repulsion","joined")=true;

	catalog<Real>("RepulsionRandom")=0.0;
	catalog<Real>("RepulsionRandom","high")=0.1;
	catalog<String>("RepulsionRandom","label")="+-";
	catalog<String>("RepulsionRandom","page")="Behavior";

	catalog<Real>("Falloff")=1.0;
	catalog<Real>("Falloff","high")=4.0;
	catalog<Real>("Falloff","max")=100.0;
	catalog<String>("Falloff","page")="Behavior";
	catalog<String>("Falloff","hint")=
			"Fade repulsion to zero over this distance.";
	catalog<bool>("Falloff","joined")=true;

	catalog<Real>("FalloffRandom")=0.0;
	catalog<Real>("FalloffRandom","high")=1.0;
	catalog<String>("FalloffRandom","label")="+-";
	catalog<String>("FalloffRandom","page")="Behavior";

	catalog<Real>("Decompression")=1.0;
	catalog<Real>("Decompression","high")=1.0;
	catalog<Real>("Decompression","max")=10.0;
	catalog<String>("Decompression","page")="Behavior";
	catalog<String>("Decompression","hint")=
			"Restore shrunken segments by pushing aside.";
	catalog<bool>("Decompression","joined")=true;

	catalog<Real>("DecompressionRandom")=0.0;
	catalog<Real>("DecompressionRandom","high")=1.0;
	catalog<String>("DecompressionRandom","label")="+-";
	catalog<String>("DecompressionRandom","page")="Behavior";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Collider Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void DodgeOp::handle(Record& a_rSignal)
{
#if FE_DODGE_DEBUG
	feLog("DodgeOp::handle\n");
#endif

	const String group=catalog<String>("Group");
	const I32 refinement=catalog<I32>("Refinement");
	const BWORD temporal=catalog<bool>("Temporal");
	const Real response=catalog<Real>("Response");
	const Real responseRandom=catalog<Real>("ResponseRandom");

	const BWORD doTemporal=temporal && (response<1.0 || responseRandom>0.0);

	m_repulsionScale=doTemporal? response: 1.0;

	//* output curves is copy of input curves
	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	if(!access(m_spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,m_spInputAccessible,
			e_primitiveGroup,group)) return;

	const U32 primitiveCount=spInputVertices->count();
	if(!primitiveCount)
	{
		feLog("spInputVertices zero count\n");
		return;
	}

	sp<SurfaceAccessorI> spInputPoints;
	if(!access(spInputPoints,m_spInputAccessible,e_point,e_position)) return;

	if(!access(m_spCollider,"Collider Surface",e_warning)) return;

	m_spCollider->setRefinement(refinement);
	m_spCollider->prepareForSearch();

	//* access DrawI interpreter
	const BWORD debug=catalog<bool>("DebugOutput");
	if(debug)
	{
		if(!accessGuide(m_spDrawDebug,a_rSignal)) return;
	}
	else
	{
		m_spDrawDebug=NULL;
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	const U32 pointCount=spInputPoints->count();
	m_pointCache.resize(pointCount);

	const I32 frame=currentFrame(a_rSignal);
	const BWORD replaced=catalog<bool>("Output Surface","replaced");
	const BWORD reuseCache=(frame==m_frame+1 || (frame==m_frame && !replaced));

#if FE_DODGE_DEBUG
	feLog("DodgeOp::handle frame %d replaced %d\n",frame,replaced);
#endif

	//* presuming always replaced with frame change
	if(doTemporal && reuseCache)
	{
#if FE_DODGE_DEBUG
		if(frame==m_frame+1)
		{
			feLog("DodgeOp::handle temporal step\n");
		}
		else
		{
			feLog("DodgeOp::handle temporal repeat\n");
		}
#endif

		sp<SurfaceAccessorI> spOutputVertices;
		if(!access(spOutputVertices,spOutputAccessible,
				e_primitiveGroup,group)) return;

		const I32 seed=catalog<I32>("Seed");

		//* output = history:prior blended with pure input
		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
				primitiveIndex++)
		{
			U32 oneSeed=seed+primitiveCount+primitiveIndex;
			const Real fraction=response+
					responseRandom*randomRealReentrant(&oneSeed,-1.0,1.0);
			if(fraction<=0.0 && !replaced)
			{
				continue;
			}
			const U32 subCount=spOutputVertices->subCount(primitiveIndex);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spInputVertices->integer(primitiveIndex,subIndex);

				SpatialVector& rCachePoint=m_pointCache[pointIndex];

				SpatialVector point=
						spInputPoints->spatialVector(pointIndex);

				if(fraction<1.0)
				{
					point=point*fraction+rCachePoint*(1.0-fraction);
				}

				rCachePoint=point;
			}
		}
	}
	else
	{
#if FE_DODGE_DEBUG
		feLog("DodgeOp::handle rebuild\n");
#endif
		//* just use new output
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector& rPoint=
					spInputPoints->spatialVector(pointIndex);
			m_pointCache[pointIndex]=rPoint;
		}
	}
	m_frame=frame;

	setNonAtomicRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

	//* flush result to output

	sp<SurfaceAccessorI> spOutputPoints;
	if(!access(spOutputPoints,spOutputAccessible,
			e_point,e_position)) return;

	for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		spOutputPoints->set(pointIndex,m_pointCache[pointIndex]);
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->popDrawMode();
	}

#if FE_DODGE_DEBUG
	feLog("DodgeOp::handle done\n");
#endif
}

void DodgeOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const Color blue(0.0,0.0,1.0);
	const Color yellow(1.0,1.0,0.0);

	const String group=catalog<String>("Group");

	const I32 seed=catalog<I32>("Seed");

	const Real threshold=catalog<Real>("Threshold");
	const Real thresholdRandom=catalog<Real>("ThresholdRandom");
	const Real projection=catalog<Real>("Projection");
	const Real projectionRandom=catalog<Real>("ProjectionRandom");
	const Real repulsion=catalog<Real>("Repulsion");
	const Real repulsionRandom=catalog<Real>("RepulsionRandom");
	const Real falloff=catalog<Real>("Falloff");
	const Real falloffRandom=catalog<Real>("FalloffRandom");
	const Real decompression=catalog<Real>("Decompression");
	const Real decompressionRandom=catalog<Real>("DecompressionRandom");

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,m_spInputAccessible,
			e_primitiveGroup,group)) return;

	Real radius=m_spCollider->radius();
	SpatialVector center=m_spCollider->center();
	const Real extendedRadius=radius+threshold+falloff;
	const Real extendedRadius2=extendedRadius*extendedRadius;

#if FE_DODGE_DEBUG
	feLog("DodgeOp::run running thread %d %s\n",
			a_id,a_spRange->brief().c_str());
	feLog("  collider center %s radius %.6G\n",c_print(center),radius);
#endif

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();
		const U32 subCount=spInputVertices->subCount(index);

		if(!subCount)
		{
			continue;
		}

		U32 oneSeed=seed+index;
		const Real oneThreshold=threshold+
				thresholdRandom*randomRealReentrant(&oneSeed,-1.0,1.0);
		const Real oneProjection=projection+
				projectionRandom*randomRealReentrant(&oneSeed,-1.0,1.0);
		const Real oneRepulsion=m_repulsionScale*(repulsion+
				repulsionRandom*randomRealReentrant(&oneSeed,-1.0,1.0));
		const Real oneFalloff=falloff+
				falloffRandom*randomRealReentrant(&oneSeed,-1.0,1.0);
		const Real oneDecompression=decompression+
				decompressionRandom*randomRealReentrant(&oneSeed,-1.0,1.0);

		//* push curve points out of collider in direction of nearest normal

		BWORD hit=FALSE;
		SpatialVector pivotPoint=spInputVertices->spatialVector(index,0);
		SpatialQuaternion pivotQuat;

		const SpatialVector rootPoint=
				spInputVertices->spatialVector(index,0);
		const SpatialVector tipPoint=
				spInputVertices->spatialVector(index,subCount-1);

		const Real approxLength=magnitude(tipPoint-rootPoint);
		const Real scopeRadius=extendedRadius+approxLength;
		const Real scopeRadius2=scopeRadius*scopeRadius;

		if(magnitudeSquared(rootPoint-center)>scopeRadius2)
		{
			continue;
		}

		//* root point
		I32 lastIndex=spInputVertices->integer(index,0);
		SpatialVector lastPoint=m_pointCache[lastIndex];

		//* NOTE skip root point
		for(U32 subIndex=1;subIndex<subCount;subIndex++)
		{
			BWORD changed=FALSE;

			const I32 pointIndex=spInputVertices->integer(index,subIndex);
			SpatialVector point=m_pointCache[pointIndex];
			const SpatialVector origPoint=point;

			SpatialVector arm1=point-pivotPoint;
			SpatialVector arm2;
/*
			if(hit)
			{
				const Real pivotScale=1.0+(oneProjection-1.0)*
						magnitude(arm1)/magnitude(tipPoint-pivotPoint);

				rotateVector(pivotQuat*pivotScale,arm1,arm2);
				point=pivotPoint+arm2;
			}
*/

#if FE_DODGE_IMPACT_DEBUG
			feLog("\nDodgeOp::handle %d %d/%d  %s\n",index,
					subIndex,subCount,c_print(point));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
#endif

			const Real mag2=magnitudeSquared(point-center);
			if(!hit && mag2>extendedRadius2)
			{
				if(mag2>scopeRadius2)
				{
					subIndex=subCount;
				}
				continue;
			}

			sp<SurfaceI::ImpactI> spImpact=m_spCollider->nearestPoint(point);
			if(spImpact.isValid())
			{
				const SpatialVector intersection=spImpact->intersection();
				const SpatialVector normal=spImpact->normal();
				const SpatialVector contact=intersection+oneThreshold*normal;
				const SpatialVector reach=contact+oneFalloff*normal;
#if FE_DODGE_IMPACT_DEBUG
				feLog("DodgeOp::handle impact %s\n",
						c_print(intersection));
#endif

				const SpatialVector to_contact=contact-point;
				const SpatialVector to_reach=reach-point;
				const Real len2=magnitudeSquared(to_reach);

				if(len2>1e-6)
				{
					Real len=sqrt(len2);

					const SpatialVector unitReach=to_reach*(1.0/len);
					Real impactDot=dot(unitReach,normal);
					if(impactDot>0.0)
					{
						const SpatialVector segmentVector=point-lastPoint;
						const Real origLength=magnitude(segmentVector);
						const SpatialVector unitSegment=
								segmentVector/(origLength+1e-6);
						const SpatialVector alongVector=
								unitSegment*dot(unitSegment,normal);
						const SpatialVector sideVector=
								unitSafe(normal-alongVector);
						SpatialVector unitRepulsion=sideVector;

						SpatialVector correction=to_contact;
						Real scalarRepulsion=oneRepulsion;

						BWORD draw=FALSE;
						if(oneFalloff>0.0 && len<oneFalloff)
						{
							//* falloff zone outside collider
							set(correction);
							const Real ratio=len/oneFalloff;
							scalarRepulsion*=ratio;
							unitRepulsion=unitSafe(sideVector*ratio+
									normal*(1.0-ratio));
						}
						else
						{
							//* inside collider
							const Real alongRepulsion=
									dot(correction,unitRepulsion);

//							scalarRepulsion-=impactDot*magnitude(correction);
							scalarRepulsion-=alongRepulsion;

							draw=TRUE;
						}

						if(scalarRepulsion>0.0)
						{
							//* excess repulsion can cause temporal bouncing
							if(scalarRepulsion>len)
							{
								scalarRepulsion=len;
							}

							correction+=scalarRepulsion*unitRepulsion;
						}

//						correction*=pow(impactDot,0.1);
						point+=correction;

						const Real newLength=magnitude(point-lastPoint);
						if(origLength>newLength)
						{
							point+=oneDecompression*(origLength-newLength)*
									sideVector;
						}

						if(m_spDrawDebug.isValid() && draw)
						{
							SpatialVector line[2];
							line[0]=point;
							line[1]=point+unitRepulsion;

							Color colors[2];
							colors[0]=blue;
							colors[1]=yellow;

							SAFEGUARD;
							m_spDrawDebug->drawLines(line,NULL,2,
									DrawI::e_discrete,false,colors);
						}

						SpatialVector newPivotPoint=m_pointCache[lastIndex];
						if(magnitudeSquared(point-newPivotPoint)>1e-4)
						{
							arm1+=pivotPoint;
							pivotPoint=newPivotPoint;
							arm1-=pivotPoint;
							changed=TRUE;
						}
					}
				}
			}

			arm2=point-pivotPoint;

#if FE_DODGE_IMPACT_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
			feLog("point %s -> %s\n",
					c_print(m_pointCache[pointIndex]),c_print(point));
#endif

			m_pointCache[pointIndex]=point;

			if(changed)
			{
				set(pivotQuat,unitSafe(arm1),unitSafe(arm2));
				hit=TRUE;

				for(U32 subIndex2=subIndex+1;subIndex2<subCount;subIndex2++)
				{
					const I32 reIndex=
							spInputVertices->integer(index,subIndex2);
					SpatialVector repoint=m_pointCache[reIndex];

					const SpatialVector rearm1=repoint-pivotPoint;
					const Real pivotScale=1.0+(oneProjection-1.0)*
							magnitude(rearm1)/magnitude(tipPoint-pivotPoint);

					SpatialVector rearm2;
					rotateVector(pivotQuat*pivotScale,rearm1,rearm2);
					repoint=pivotPoint+rearm2;

					m_pointCache[reIndex]=repoint;
				}
			}
#if FE_DODGE_IMPACT_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
#endif
			lastPoint=origPoint;
			lastIndex=pointIndex;
		}
	}
}
